@@ -15,8 +15,10 @@ import functools
 import hashlib
 import json
 import logging
+import os
 import platform
 import socket
+import sys
 import time
 import uuid
 
@@ -27,6 +29,7 @@ from six.moves import urllib
 
 import keystoneauth1
 from keystoneauth1 import _utils as utils
+from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 
 try:
@@ -43,6 +46,11 @@ DEFAULT_USER_AGENT = 'keystoneauth1/%s %s %s/%s' % (
     keystoneauth1.__version__, requests.utils.default_user_agent(),
     platform.python_implementation(), platform.python_version())
 
+# NOTE(jamielennox): Clients will likely want to print more than json. Please
+# propose a patch if you have a content type you think is reasonable to print
+# here and we'll add it to the list as required.
+_LOG_CONTENT_TYPES = set(['application/json'])
+
 _logger = utils.get_logger(__name__)
 
 
@@ -58,6 +66,22 @@ def _construct_session(session_obj=None):
     return session_obj
 
 
+def _mv_legacy_headers_for_service(mv_service_type):
+    """Workaround for services that predate standardization.
+
+    TODO(sdague): eventually convert this to using os-service-types
+    and put the logic there. However, right now this is so little
+    logic, inlining it for release is a better call.
+
+    """
+    headers = []
+    if mv_service_type == "compute":
+        headers.append("X-OpenStack-Nova-API-Version")
+    elif mv_service_type == "baremetal":
+        headers.append("X-OpenStack-Ironic-API-Version")
+    return headers
+
+
 class _JSONEncoder(json.JSONEncoder):
 
     def default(self, o):
@@ -72,7 +96,7 @@ class _JSONEncoder(json.JSONEncoder):
 
 
 class _StringFormatter(object):
-    """A String formatter that fetches values on demand"""
+    """A String formatter that fetches values on demand."""
 
     def __init__(self, session, auth):
         self.session = session
@@ -93,20 +117,89 @@ class _StringFormatter(object):
         return value
 
 
+def _determine_calling_package():
+    """Walk the call frames trying to identify what is using this module."""
+    # Create a lookup table mapping file name to module name. The ``inspect``
+    # module does this but is far less efficient. Same story with the
+    # frame walking below.  One could use ``inspect.stack()`` but it
+    # has far more overhead.
+    mod_lookup = dict((m.__file__, n) for n, m in sys.modules.items()
+                      if hasattr(m, '__file__'))
+
+    # NOTE(shaleh): these are not useful because they hide the real
+    # user of the code. debtcollector did not import keystoneauth but
+    # it will show up in the call stack. Similarly we do not want to
+    # report ourselves or keystone client as the user agent. The real
+    # user is the code importing them.
+    ignored = ('debtcollector', 'keystoneauth1', 'keystoneclient')
+
+    i = 0
+    while True:
+        i += 1
+
+        try:
+            # NOTE(shaleh): this is safe in CPython but could break in
+            # other implementations of Python. Yes, the `inspect`
+            # module could be used instead. But it does a lot more
+            # work so it has worse performance.
+            f = sys._getframe(i)
+            try:
+                name = mod_lookup[f.f_code.co_filename]
+                # finds the full name module.foo.bar but all we need
+                # is the module name.
+                name, _, _ = name.partition('.')
+                if name not in ignored:
+                    return name
+            except KeyError:
+                pass  # builtin or the like
+        except ValueError:
+            # hit the bottom of the frame stack
+            break
+
+    return ''
+
+
+def _determine_user_agent():
+    """Attempt to programatically generate a user agent string.
+
+    First, look at the name of the process. Return this unless it is in
+    the `ignored` list.  Otherwise, look at the function call stack and
+    try to find the name of the code that invoked this module.
+    """
+    # NOTE(shaleh): mod_wsgi is not any more useful than just
+    # reporting "keystoneauth". Ignore it and perform the package name
+    # heuristic.
+    ignored = ('mod_wsgi', )
+
+    try:
+        name = sys.argv[0]
+    except IndexError:
+        # sys.argv is empty, usually the Python interpreter prevents this.
+        return ''
+
+    if not name:
+        return ''
+
+    name = os.path.basename(name)
+    if name in ignored:
+        name = _determine_calling_package()
+    return name
+
+
 class Session(object):
     """Maintains client communication state and common functionality.
 
     As much as possible the parameters to this class reflect and are passed
-    directly to the requests library.
+    directly to the :mod:`requests` library.
 
     :param auth: An authentication plugin to authenticate the session with.
                  (optional, defaults to None)
-    :type auth: :py:class:`keystonauth.auth.base.BaseAuthPlugin`
+    :type auth: keystoneauth1.plugin.BaseAuthPlugin
     :param requests.Session session: A requests session object that can be used
                                      for issuing requests. (optional)
-    :param string original_ip: The original IP of the requesting user which
-                               will be sent to identity service in a
-                               'Forwarded' header. (optional)
+    :param str original_ip: The original IP of the requesting user which will
+                            be sent to identity service in a 'Forwarded'
+                            header. (optional)
     :param verify: The verification arguments to pass to requests. These are of
                    the same form as requests expects, so True or False to
                    verify (or not) against system certificates or a path to a
@@ -121,29 +214,51 @@ class Session(object):
                           numerical value indicating some amount (or fraction)
                           of seconds or 0 for no timeout. (optional, defaults
                           to 0)
-    :param string user_agent: A User-Agent header string to use for the
-                              request. If not provided, a default of
-                              :attr:`~keystoneauth1.session.DEFAULT_USER_AGENT`
-                              is used, which contains the keystoneauth1 version
-                              as well as those of the requests library and
-                              which Python is being used. When a non-None value
-                              is passed, it will be prepended to the default.
+    :param str user_agent: A User-Agent header string to use for the request.
+                           If not provided, a default of
+                           :attr:`~keystoneauth1.session.DEFAULT_USER_AGENT` is
+                           used, which contains the keystoneauth1 version as
+                           well as those of the requests library and which
+                           Python is being used. When a non-None value is
+                           passed, it will be prepended to the default.
     :param int/bool redirect: Controls the maximum number of redirections that
                               can be followed by a request. Either an integer
                               for a specific count or True/False for
                               forever/never. (optional, default to 30)
+    :param dict additional_headers: Additional headers that should be attached
+                                    to every request passing through the
+                                    session. Headers of the same name specified
+                                    per request will take priority.
+    :param str app_name: The name of the application that is creating the
+                         session. This will be used to create the user_agent.
+    :param str app_version: The version of the application creating the
+                            session. This will be used to create the
+                            user_agent.
+    :param list additional_user_agent: A list of tuple of name, version that
+                                       will be added to the user agent. This
+                                       can be used by libraries that are part
+                                       of the communication process.
+    :param dict discovery_cache: A dict to be used for caching of discovery
+                                 information. This is normally managed
+                                 transparently, but if the user wants to
+                                 share a single cache across multiple sessions
+                                 that do not share an auth plugin, it can
+                                 be provided here. (optional, defaults to
+                                 None which means automatically manage)
     """
 
     user_agent = None
 
-    _REDIRECT_STATUSES = (301, 302, 303, 305, 307)
+    _REDIRECT_STATUSES = (301, 302, 303, 305, 307, 308)
 
     _DEFAULT_REDIRECT_LIMIT = 30
 
     @positional(2)
     def __init__(self, auth=None, session=None, original_ip=None, verify=True,
                  cert=None, timeout=None, user_agent=None,
-                 redirect=_DEFAULT_REDIRECT_LIMIT):
+                 redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
+                 app_name=None, app_version=None, additional_user_agent=None,
+                 discovery_cache=None):
 
         self.auth = auth
         self.session = _construct_session(session)
@@ -152,16 +267,19 @@ class Session(object):
         self.cert = cert
         self.timeout = None
         self.redirect = redirect
+        self.additional_headers = additional_headers or {}
+        self.app_name = app_name
+        self.app_version = app_version
+        self.additional_user_agent = additional_user_agent or []
+        self._determined_user_agent = None
+        if discovery_cache is None:
+            discovery_cache = {}
+        self._discovery_cache = discovery_cache
 
         if timeout is not None:
             self.timeout = float(timeout)
 
-        # don't override the class variable if none provided
         if user_agent is not None:
-            # Per RFC 7231 Section 5.5.3, identifiers in a user-agent
-            # should be ordered by decreasing significance.
-            # If a user sets their product, we prepend it to the KSA
-            # version, requests version, and then the Python version.
             self.user_agent = "%s %s" % (user_agent, DEFAULT_USER_AGENT)
 
         self._json = _JSONEncoder()
@@ -198,9 +316,9 @@ class Session(object):
 
     @staticmethod
     def _process_header(header):
-        """Redacts the secure headers to be logged."""
+        """Redact the secure headers to be logged."""
         secure_headers = ('authorization', 'x-auth-token',
-                          'x-subject-token',)
+                          'x-subject-token', 'x-service-token')
         if header[0].lower() in secure_headers:
             token_hasher = hashlib.sha1()
             token_hasher.update(header[1].encode('utf-8'))
@@ -210,7 +328,8 @@ class Session(object):
 
     @positional()
     def _http_log_request(self, url, method=None, data=None,
-                          json=None, headers=None, logger=_logger):
+                          json=None, headers=None, query_params=None,
+                          logger=_logger):
         if not logger.isEnabledFor(logging.DEBUG):
             # NOTE(morganfainberg): This whole debug section is expensive,
             # there is no need to do the work if we're not going to emit a
@@ -229,10 +348,18 @@ class Session(object):
         if method:
             string_parts.extend(['-X', method])
 
-        string_parts.append(url)
+        if query_params:
+            # Don't check against `is not None` as this can be
+            # an empty dictionary, which we shouldn't bother logging.
+            url = url + '?' + urllib.parse.urlencode(query_params)
+            # URLs with query strings need to be wrapped in quotes in order
+            # for the CURL command to run properly.
+            string_parts.append('"%s"' % url)
+        else:
+            string_parts.append(url)
 
         if headers:
-            for header in six.iteritems(headers):
+            for header in headers.items():
                 string_parts.append('-H "%s: %s"'
                                     % self._process_header(header))
         if json:
@@ -260,7 +387,26 @@ class Session(object):
             if not headers:
                 headers = response.headers
             if not text:
-                text = self._remove_service_catalog(response.text)
+                # NOTE(samueldmq): If the response does not provide enough info
+                # about the content type to decide whether it is useful and
+                # safe to log it or not, just do not log the body. Trying to
+                # read the response body anyways may result on reading a long
+                # stream of bytes and getting an unexpected MemoryError. See
+                # bug 1616105 for further details.
+                content_type = response.headers.get('content-type', None)
+
+                # NOTE(lamt): Per [1], the Content-Type header can be of the
+                # form Content-Type := type "/" subtype *[";" parameter]
+                # [1] https://www.w3.org/Protocols/rfc1341/4_Content-Type.html
+                for log_type in _LOG_CONTENT_TYPES:
+                    if content_type is not None and content_type.startswith(
+                            log_type):
+                        text = self._remove_service_catalog(response.text)
+                        break
+                else:
+                    text = ('Omitted, Content-Type is set to %s. Only '
+                            '%s responses have their bodies logged.')
+                    text = text % (content_type, ', '.join(_LOG_CONTENT_TYPES))
         if json:
             text = self._json.encode(json)
 
@@ -269,19 +415,64 @@ class Session(object):
         if status_code:
             string_parts.append('[%s]' % status_code)
         if headers:
-            for header in six.iteritems(headers):
+            for header in headers.items():
                 string_parts.append('%s: %s' % self._process_header(header))
         if text:
             string_parts.append('\nRESP BODY: %s\n' % text)
 
         logger.debug(' '.join(string_parts))
 
+    @staticmethod
+    def _set_microversion_headers(
+            headers, microversion, service_type, endpoint_filter):
+        # We're converting it to normalized version number for two reasons.
+        # First, to validate it's a real version number. Second, so that in
+        # the future we can pre-validate that it is within the range of
+        # available microversions before we send the request.
+        # TODO(mordred) Validate when we get the response back that
+        # the server executed in the microversion we expected.
+        # TODO(mordred) Validate that the requested microversion works
+        # with the microversion range we found in discovery.
+        microversion = discover.normalize_version_number(microversion)
+        # Can't specify a M.latest microversion
+        if (microversion[0] != discover.LATEST and
+                discover.LATEST in microversion[1:]):
+            raise TypeError(
+                "Specifying a '{major}.latest' microversion is not allowed.")
+        microversion = discover.version_to_string(microversion)
+        if not service_type:
+            if endpoint_filter and 'service_type' in endpoint_filter:
+                service_type = endpoint_filter['service_type']
+            else:
+                raise TypeError(
+                    "microversion {microversion} was requested but no"
+                    " service_type information is available. Either provide a"
+                    " service_type in endpoint_filter or pass"
+                    " microversion_service_type as an argument.".format(
+                        microversion=microversion))
+
+        # TODO(mordred) cinder uses volume in its microversion header. This
+        # logic should be handled in the future by os-service-types but for
+        # now hard-code for cinder.
+        if (service_type.startswith('volume')
+                or service_type == 'block-storage'):
+            service_type = 'volume'
+        headers.setdefault('OpenStack-API-Version',
+                           '{service_type} {microversion}'.format(
+                               service_type=service_type,
+                               microversion=microversion))
+        header_names = _mv_legacy_headers_for_service(service_type)
+        for h in header_names:
+            headers.setdefault(h, microversion)
+
     @positional()
     def request(self, url, method, json=None, original_ip=None,
                 user_agent=None, redirect=None, authenticated=None,
                 endpoint_filter=None, auth=None, requests_auth=None,
                 raise_exc=True, allow_reauth=True, log=True,
                 endpoint_override=None, connect_retries=0, logger=_logger,
+                allow={}, client_name=None, client_version=None,
+                microversion=None, microversion_service_type=None,
                 **kwargs):
         """Send an HTTP request with the specified characteristics.
 
@@ -291,19 +482,18 @@ class Session(object):
         Arguments that are not handled are passed through to the requests
         library.
 
-        :param string url: Path or fully qualified URL of HTTP request. If only
-                           a path is provided then endpoint_filter must also be
-                           provided such that the base URL can be determined.
-                           If a fully qualified URL is provided then
-                           endpoint_filter will be ignored.
-        :param string method: The http method to use. (e.g. 'GET', 'POST')
-        :param string original_ip: Mark this request as forwarded for this ip.
-                                   (optional)
+        :param str url: Path or fully qualified URL of HTTP request. If only a
+                        path is provided then endpoint_filter must also be
+                        provided such that the base URL can be determined. If a
+                        fully qualified URL is provided then endpoint_filter
+                        will be ignored.
+        :param str method: The http method to use. (e.g. 'GET', 'POST')
+        :param str original_ip: Mark this request as forwarded for this ip.
+                                (optional)
         :param dict headers: Headers to be included in the request. (optional)
         :param json: Some data to be represented as JSON. (optional)
-        :param string user_agent: A user_agent to use for the request. If
-                                  present will override one present in headers.
-                                  (optional)
+        :param str user_agent: A user_agent to use for the request. If present
+                               will override one present in headers. (optional)
         :param int/bool redirect: the maximum number of redirections that
                                   can be followed by a request. Either an
                                   integer for a specific count or True/False
@@ -325,14 +515,14 @@ class Session(object):
                                       ignored if a fully qualified URL is
                                       provided but take priority over an
                                       endpoint_filter. This string may contain
-                                      the values %(project_id)s and %(user_id)s
-                                      to have those values replaced by the
-                                      project_id/user_id of the current
-                                      authentication. (optional)
+                                      the values ``%(project_id)s`` and
+                                      ``%(user_id)s`` to have those values
+                                      replaced by the project_id/user_id of the
+                                      current authentication. (optional)
         :param auth: The auth plugin to use when authenticating this request.
                      This will override the plugin that is attached to the
                      session (if any). (optional)
-        :type auth: :py:class:`keystonauth.auth.base.BaseAuthPlugin`
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
         :param requests_auth: A requests library auth plugin that cannot be
                               passed via kwarg because the `auth` kwarg
                               collides with our own auth plugins. (optional)
@@ -346,22 +536,40 @@ class Session(object):
         :param bool log: If True then log the request and response data to the
                          debug log. (optional, default True)
         :param logger: The logger object to use to log request and responses.
-                       If not provided the keystonauth.session default
+                       If not provided the keystoneauth1.session default
                        logger will be used.
         :type logger: logging.Logger
+        :param dict allow: Extra filters to pass when discovering API
+                           versions. (optional)
+        :param microversion: Microversion to send for this request.
+                       microversion can be given as a string or a tuple.
+                       (optional)
+        :param str microversion_service_type: The service_type to be sent in
+                       the microversion header, if a microversion is given.
+                       Defaults to the value of service_type from
+                       endpoint_filter if one exists. If endpoint_filter is not
+                       provided or does not have a service_type, microversion
+                       is given and microversion_service_type is not provided,
+                       an exception will be raised.
         :param kwargs: any other parameter that can be passed to
-                       requests.Session.request (such as `headers`). Except:
-                       'data' will be overwritten by the data in 'json' param.
-                       'allow_redirects' is ignored as redirects are handled
-                       by the session.
+                       :meth:`requests.Session.request` (such as `headers`).
+                       Except:
 
-        :raises keystonauth.exceptions.ClientException: For connection
+                       - `data` will be overwritten by the data in the `json`
+                         param.
+                       - `allow_redirects` is ignored as redirects are handled
+                         by the session.
+
+        :raises keystoneauth1.exceptions.base.ClientException: For connection
             failure, or to indicate an error response code.
 
         :returns: The response to the request.
         """
-
         headers = kwargs.setdefault('headers', dict())
+        if microversion:
+            self._set_microversion_headers(
+                headers, microversion, microversion_service_type,
+                endpoint_filter)
 
         if authenticated is None:
             authenticated = bool(auth or self.auth)
@@ -388,7 +596,8 @@ class Session(object):
             if endpoint_override:
                 base_url = endpoint_override % _StringFormatter(self, auth)
             elif endpoint_filter:
-                base_url = self.get_endpoint(auth, **endpoint_filter)
+                base_url = self.get_endpoint(auth, allow=allow,
+                                             **endpoint_filter)
 
             if not base_url:
                 raise exceptions.EndpointNotFound()
@@ -406,25 +615,67 @@ class Session(object):
         elif self.user_agent:
             user_agent = headers.setdefault('User-Agent', self.user_agent)
         else:
-            user_agent = headers.setdefault('User-Agent', DEFAULT_USER_AGENT)
+            # Per RFC 7231 Section 5.5.3, identifiers in a user-agent should be
+            # ordered by decreasing significance.  If a user sets their product
+            # that value will be used. Otherwise we attempt to derive a useful
+            # product value. The value will be prepended it to the KSA version,
+            # requests version, and then the Python version.
+
+            agent = []
+
+            if self.app_name and self.app_version:
+                agent.append('%s/%s' % (self.app_name, self.app_version))
+            elif self.app_name:
+                agent.append(self.app_name)
+
+            for additional in self.additional_user_agent:
+                agent.append('%s/%s' % additional)
+
+            if client_name and client_version:
+                agent.append('%s/%s' % (client_name, client_version))
+            elif client_name:
+                agent.append(client_name)
+
+            if not agent:
+                # NOTE(jamielennox): determine_user_agent will return an empty
+                # string on failure so checking for None will ensure it is only
+                # called once even on failure.
+                if self._determined_user_agent is None:
+                    self._determined_user_agent = _determine_user_agent()
+
+                if self._determined_user_agent:
+                    agent.append(self._determined_user_agent)
+
+            agent.append(DEFAULT_USER_AGENT)
+            user_agent = headers.setdefault('User-Agent', ' '.join(agent))
 
         if self.original_ip:
             headers.setdefault('Forwarded',
                                'for=%s;by=%s' % (self.original_ip, user_agent))
 
         if json is not None:
-            headers['Content-Type'] = 'application/json'
+            headers.setdefault('Content-Type', 'application/json')
             kwargs['data'] = self._json.encode(json)
 
+        for k, v in self.additional_headers.items():
+            headers.setdefault(k, v)
+
         kwargs.setdefault('verify', self.verify)
 
         if requests_auth:
             kwargs['auth'] = requests_auth
 
+        # Query parameters that are included in the url string will
+        # be logged properly, but those sent in the `params` parameter
+        # (which the requests library handles) need to be explicitly
+        # picked out so they can be included in the URL that gets loggged.
+        query_params = kwargs.get('params', dict())
+
         if log:
             self._http_log_request(url, method=method,
                                    data=kwargs.get('data'),
                                    headers=headers,
+                                   query_params=query_params,
                                    logger=logger)
 
         # Force disable requests redirect handling. We will manage this below.
@@ -451,6 +702,30 @@ class Session(object):
 
         resp = send(**kwargs)
 
+        # log callee and caller request-id for each api call
+        if log:
+            # service_name should be fetched from endpoint_filter if it is not
+            # present then use service_type as service_name.
+            service_name = None
+            if endpoint_filter:
+                service_name = endpoint_filter.get('service_name')
+                if not service_name:
+                    service_name = endpoint_filter.get('service_type')
+
+            # Nova uses 'x-compute-request-id' and other services like
+            # Glance, Cinder etc are using 'x-openstack-request-id' to store
+            # request-id in the header
+            request_id = (resp.headers.get('x-openstack-request-id') or
+                          resp.headers.get('x-compute-request-id'))
+            if request_id:
+                logger.debug('%(method)s call to %(service_name)s for '
+                             '%(url)s used request id '
+                             '%(response_request_id)s',
+                             {'method': resp.request.method,
+                              'service_name': service_name,
+                              'url': resp.url,
+                              'response_request_id': request_id})
+
         # handle getting a 401 Unauthorized response by invalidating the plugin
         # and then retrying the request. This is only tried once.
         if resp.status_code == 401 and authenticated and allow_reauth:
@@ -491,8 +766,14 @@ class Session(object):
             except requests.exceptions.Timeout:
                 msg = 'Request to %s timed out' % url
                 raise exceptions.ConnectTimeout(msg)
-            except requests.exceptions.ConnectionError:
-                msg = 'Unable to establish connection to %s' % url
+            except requests.exceptions.ConnectionError as e:
+                # NOTE(sdague): urllib3/requests connection error is a
+                # translation of SocketError. However, SocketError
+                # happens for many different reasons, and that low
+                # level message is often really important in figuring
+                # out the difference between network misconfigurations
+                # and firewall blocking.
+                msg = 'Unable to establish connection to %s: %s' % (url, e)
                 raise exceptions.ConnectFailure(msg)
             except requests.exceptions.RequestException as e:
                 msg = 'Unexpected exception for %(url)s: %(error)s' % {
@@ -610,15 +891,15 @@ class Session(object):
 
         :param auth: The auth plugin to use for token. Overrides the plugin
                      on the session. (optional)
-        :type auth: :py:class:`keystonauth.auth.base.BaseAuthPlugin`
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
 
-        :raises keystonauth.exceptions.AuthorizationFailure: if a new token
-                                                                fetch fails.
-        :raises keystonauth.exceptions.MissingAuthPlugin: if a plugin is not
-                                                             available.
+        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
+            if a new token fetch fails.
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
+            if a plugin is not available.
 
         :returns: Authentication headers or None for failure.
-        :rtype: dict
+        :rtype: :class:`dict`
         """
         auth = self._auth_required(auth, 'fetch a token')
         return auth.get_headers(self, **kwargs)
@@ -628,16 +909,17 @@ class Session(object):
 
         :param auth: The auth plugin to use for token. Overrides the plugin
                      on the session. (optional)
-        :type auth: :py:class:`keystonauth.auth.base.BaseAuthPlugin`
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
 
-        :raises keystonauth.exceptions.AuthorizationFailure: if a new token
-                                                                fetch fails.
-        :raises keystonauth.exceptions.MissingAuthPlugin: if a plugin is not
-                                                             available.
+        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
+             if a new token fetch fails.
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
+            if a plugin is not available.
 
-        *DEPRECATED*: This assumes that the only header that is used to
-                      authenticate a message is 'X-Auth-Token'. This may not be
-                      correct. Use get_auth_headers instead.
+        .. warning::
+            **DEPRECATED**: This assumes that the only header that is used to
+            authenticate a message is ``X-Auth-Token``. This may not be
+            correct. Use :meth:`get_auth_headers` instead.
 
         :returns: A valid token.
         :rtype: string
@@ -649,17 +931,38 @@ class Session(object):
 
         :param auth: The auth plugin to use for token. Overrides the plugin on
                      the session. (optional)
-        :type auth: :py:class:`keystonauth.auth.base.BaseAuthPlugin`
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
 
-        :raises keystonauth.exceptions.MissingAuthPlugin: if a plugin is not
-                                                             available.
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
+            if a plugin is not available.
 
         :returns: An endpoint if available or None.
         :rtype: string
         """
+        if 'endpoint_override' in kwargs:
+            return kwargs['endpoint_override']
+
         auth = self._auth_required(auth, 'determine endpoint URL')
+
         return auth.get_endpoint(self, **kwargs)
 
+    def get_endpoint_data(self, auth=None, **kwargs):
+        """Get endpoint data as provided by the auth plugin.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
+            if a plugin is not available.
+        :raises TypeError: If arguments are invalid
+
+        :returns: Endpoint data if available or None.
+        :rtype: keystoneauth1.discover.EndpointData
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_endpoint_data(self, **kwargs)
+
     def get_auth_connection_params(self, auth=None, **kwargs):
         """Return auth connection params as provided by the auth plugin.
 
@@ -680,20 +983,20 @@ class Session(object):
 
         :param auth: The auth plugin to use for tokens. Overrides the plugin
                      on the session. (optional)
-        :type auth: keystoneclient.auth.base.BaseAuthPlugin
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
 
-        :raises keystoneclient.exceptions.AuthorizationFailure: if a new token
-                                                                fetch fails.
-        :raises keystoneclient.exceptions.MissingAuthPlugin: if a plugin is not
-                                                             available.
-        :raises keystoneclient.exceptions.UnsupportedParameters: if the plugin
-            returns a parameter that is not supported by this session.
+        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
+            if a new token fetch fails.
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
+            if a plugin is not available.
+        :raises keystoneauth1.exceptions.auth_plugins.UnsupportedParameters:
+            if the plugin returns a parameter that is not supported by this
+            session.
 
         :returns: Authentication headers or None for failure.
-        :rtype: dict
+        :rtype: :class:`dict`
         """
-        msg = 'An auth plugin is required to fetch connection params'
-        auth = self._auth_required(auth, msg)
+        auth = self._auth_required(auth, 'fetch connection params')
         params = auth.get_connection_params(self, **kwargs)
 
         # NOTE(jamielennox): There needs to be some consensus on what
@@ -718,7 +1021,7 @@ class Session(object):
 
         :param auth: The auth plugin to invalidate. Overrides the plugin on the
                      session. (optional)
-        :type auth: :py:class:`keystonauth.auth.base.BaseAuthPlugin`
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
 
         """
         auth = self._auth_required(auth, 'validate')
@@ -729,14 +1032,15 @@ class Session(object):
 
         :param auth: The auth plugin to use for token. Overrides the plugin
                      on the session. (optional)
-        :type auth: keystonauth.auth.base.BaseAuthPlugin
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
 
-        :raises keystonauth.exceptions.AuthorizationFailure:
+        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
             if a new token fetch fails.
-        :raises keystonauth.exceptions.MissingAuthPlugin:
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
             if a plugin is not available.
 
-        :returns string: Current user_id or None if not supported by plugin.
+        :returns: Current user_id or None if not supported by plugin.
+        :rtype: :class:`str`
         """
         auth = self._auth_required(auth, 'get user_id')
         return auth.get_user_id(self)
@@ -746,14 +1050,15 @@ class Session(object):
 
         :param auth: The auth plugin to use for token. Overrides the plugin
                      on the session. (optional)
-        :type auth: keystonauth.auth.base.BaseAuthPlugin
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
 
-        :raises keystonauth.exceptions.AuthorizationFailure:
+        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
             if a new token fetch fails.
-        :raises keystonauth.exceptions.MissingAuthPlugin:
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
             if a plugin is not available.
 
-        :returns string: Current project_id or None if not supported by plugin.
+        :returns: Current project_id or None if not supported by plugin.
+        :rtype: :class:`str`
         """
         auth = self._auth_required(auth, 'get project_id')
         return auth.get_project_id(self)
@@ -767,8 +1072,9 @@ class TCPKeepAliveAdapter(requests.adapters.HTTPAdapter):
 
     This Adapter also preserves the default behaviour of Requests which
     disables Nagle's Algorithm. See also:
-    http://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
+    https://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
     """
+
     def init_poolmanager(self, *args, **kwargs):
         if 'socket_options' not in kwargs and REQUESTS_VERSION >= (2, 4, 1):
             socket_options = [
@@ -786,7 +1092,9 @@ class TCPKeepAliveAdapter(requests.adapters.HTTPAdapter):
                     (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 60)
                 ]
 
-            if hasattr(socket, 'TCP_KEEPCNT'):
+            # Windows subsystem for Linux does not support this feature
+            if (hasattr(socket, 'TCP_KEEPCNT') and
+                    not utils.is_windows_linux_subsystem):
                 socket_options += [
                     # Set the maximum number of keep-alive probes
                     (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 4),
